# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper around debugger plugin to conditionally enable it."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

import six
from werkzeug import wrappers

from tensorboard.backend import http_util
from tensorboard.plugins import base_plugin
from tensorboard.plugins.debugger import constants
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()


class InactiveDebuggerPlugin(base_plugin.TBPlugin):
    """A placeholder debugger plugin used when no grpc port is specified."""

    plugin_name = constants.DEBUGGER_PLUGIN_NAME

    def __init__(self):
        pass

    def is_active(self):
        return False

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(
            element_name="tf-debugger-dashboard"
        )

    def get_plugin_apps(self):
        return {
            "/debugger_grpc_host_port": self._serve_debugger_grpc_host_port,
        }

    @wrappers.Request.application
    def _serve_debugger_grpc_host_port(self, request):
        # Respond with a -1 port number to indicate the debugger plugin is
        # inactive.
        return http_util.Respond(
            request, {"host": None, "port": -1}, "application/json"
        )


class DebuggerPluginLoader(base_plugin.TBLoader):
    """DebuggerPlugin factory factory.

    This class determines which debugger plugin to load, based on custom
    flags. It also checks for the `grpcio` PyPi dependency.
    """

    def define_flags(self, parser):
        """Adds DebuggerPlugin CLI flags to parser."""
        group = parser.add_argument_group("debugger plugin")
        group.add_argument(
            "--debugger_data_server_grpc_port",
            metavar="PORT",
            type=int,
            default=-1,
            help="""\
The port at which the non-interactive debugger data server should
receive debugging data via gRPC from one or more debugger-enabled
TensorFlow runtimes. No debugger plugin or debugger data server will be
started if this flag is not provided. This flag differs from the
`--debugger_port` flag in that it starts a non-interactive mode. It is
for use with the "health pills" feature of the Graph Dashboard. This
flag is mutually exclusive with `--debugger_port`.\
""",
        )
        group.add_argument(
            "--debugger_port",
            metavar="PORT",
            type=int,
            default=-1,
            help="""\
The port at which the interactive debugger data server (to be started by
the debugger plugin) should receive debugging data via gRPC from one or
more debugger-enabled TensorFlow runtimes. No debugger plugin or
debugger data server will be started if this flag is not provided. This
flag differs from the `--debugger_data_server_grpc_port` flag in that it
starts an interactive mode that allows user to pause at selected nodes
inside a TensorFlow Graph or between Session.runs. It is for use with
the interactive Debugger Dashboard. This flag is mutually exclusive with
`--debugger_data_server_grpc_port`.\
""",
        )

    def fix_flags(self, flags):
        """Fixes Debugger related flags.

        Raises:
          ValueError: If both the `debugger_data_server_grpc_port` and
            `debugger_port` flags are specified as >= 0.
        """
        # Check that not both grpc port flags are specified.
        if flags.debugger_data_server_grpc_port > 0 and flags.debugger_port > 0:
            raise base_plugin.FlagsError(
                "--debugger_data_server_grpc_port and --debugger_port are mutually "
                "exclusive. Do not use both of them at the same time."
            )

    def load(self, context):
        """Returns the debugger plugin, if possible.

        Args:
          context: The TBContext flags including `add_arguments`.

        Returns:
          A DebuggerPlugin instance or None if it couldn't be loaded.
        """
        flags = context.flags
        if flags.debugger_data_server_grpc_port > 0 or flags.debugger_port > 0:
            # Verify that the required Python packages are installed.
            try:
                # pylint: disable=unused-import
                import tensorflow
            except ImportError:
                raise ImportError(
                    "To use the debugger plugin, you need to have TensorFlow installed:\n"
                    "  pip install tensorflow"
                )

        if flags.debugger_data_server_grpc_port > 0:
            from tensorboard.plugins.debugger import (
                debugger_plugin as debugger_plugin_lib,
            )

            # debugger_data_server_grpc opens the non-interactive Debugger Plugin,
            # which appears as health pills in the Graph Plugin.
            noninteractive_plugin = debugger_plugin_lib.DebuggerPlugin(context)
            logger.info(
                "Starting Non-interactive Debugger Plugin at gRPC port %d",
                flags.debugger_data_server_grpc_port,
            )
            noninteractive_plugin.listen(flags.debugger_data_server_grpc_port)
            return noninteractive_plugin
        elif flags.debugger_port > 0:
            from tensorboard.plugins.debugger import (
                interactive_debugger_plugin as interactive_debugger_plugin_lib,
            )

            interactive_plugin = interactive_debugger_plugin_lib.InteractiveDebuggerPlugin(
                context
            )
            logger.info(
                "Starting Interactive Debugger Plugin at gRPC port %d",
                flags.debugger_data_server_grpc_port,
            )
            interactive_plugin.listen(flags.debugger_port)
            return interactive_plugin
        else:
            # If neither the debugger_data_server_grpc_port flag or the grpc_port
            # flag is specified, we instantiate a dummy plugin as a placeholder for
            # the frontend. The dummy plugin will display a message indicating that
            # the plugin is not active. It'll also display a command snippet to
            # illustrate how to activate the interactive Debugger Plugin.
            return InactiveDebuggerPlugin()



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""TensorBoard is a webapp for understanding TensorFlow runs and graphs."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import threading
import types


def lazy_load(name):
    """Decorator to define a function that lazily loads the module 'name'.

    This can be used to defer importing troublesome dependencies - e.g. ones that
    are large and infrequently used, or that cause a dependency cycle -
    until they are actually used.

    Args:
      name: the fully-qualified name of the module; typically the last segment
        of 'name' matches the name of the decorated function

    Returns:
      Decorator function that produces a lazy-loading module 'name' backed by the
      underlying decorated function.
    """

    def wrapper(load_fn):
        # Wrap load_fn to call it exactly once and update __dict__ afterwards to
        # make future lookups efficient (only failed lookups call __getattr__).
        @_memoize
        def load_once(self):
            if load_once.loading:
                raise ImportError(
                    "Circular import when resolving LazyModule %r" % name
                )
            load_once.loading = True
            try:
                module = load_fn()
            finally:
                load_once.loading = False
            self.__dict__.update(module.__dict__)
            load_once.loaded = True
            return module

        load_once.loading = False
        load_once.loaded = False

        # Define a module that proxies getattr() and dir() to the result of calling
        # load_once() the first time it's needed. The class is nested so we can close
        # over load_once() and avoid polluting the module's attrs with our own state.
        class LazyModule(types.ModuleType):
            def __getattr__(self, attr_name):
                return getattr(load_once(self), attr_name)

            def __dir__(self):
                return dir(load_once(self))

            def __repr__(self):
                if load_once.loaded:
                    return "<%r via LazyModule (loaded)>" % load_once(self)
                return (
                    "<module %r via LazyModule (not yet loaded)>"
                    % self.__name__
                )

        return LazyModule(name)

    return wrapper


def _memoize(f):
    """Memoizing decorator for f, which must have exactly 1 hashable
    argument."""
    nothing = object()  # Unique "no value" sentinel object.
    cache = {}
    # Use a reentrant lock so that if f references the resulting wrapper we die
    # with recursion depth exceeded instead of deadlocking.
    lock = threading.RLock()

    @functools.wraps(f)
    def wrapper(arg):
        if cache.get(arg, nothing) is nothing:
            with lock:
                if cache.get(arg, nothing) is nothing:
                    cache[arg] = f(arg)
        return cache[arg]

    return wrapper



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Information about histogram summaries."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.compat.proto import summary_pb2
from tensorboard.plugins.histogram import plugin_data_pb2
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

PLUGIN_NAME = "histograms"

# The most recent value for the `version` field of the
# `HistogramPluginData` proto.
PROTO_VERSION = 0


def create_summary_metadata(display_name, description):
    """Create a `summary_pb2.SummaryMetadata` proto for histogram plugin data.

    Returns:
      A `summary_pb2.SummaryMetadata` protobuf object.
    """
    content = plugin_data_pb2.HistogramPluginData(version=PROTO_VERSION)
    return summary_pb2.SummaryMetadata(
        display_name=display_name,
        summary_description=description,
        plugin_data=summary_pb2.SummaryMetadata.PluginData(
            plugin_name=PLUGIN_NAME, content=content.SerializeToString()
        ),
    )


def parse_plugin_metadata(content):
    """Parse summary metadata to a Python object.

    Arguments:
      content: The `content` field of a `SummaryMetadata` proto
        corresponding to the histogram plugin.

    Returns:
      A `HistogramPluginData` protobuf object.
    """
    if not isinstance(content, bytes):
        raise TypeError("Content type must be bytes")
    if content == b"{}":
        # Old-style JSON format. Equivalent to an all-default proto.
        return plugin_data_pb2.HistogramPluginData()
    else:
        result = plugin_data_pb2.HistogramPluginData.FromString(content)
        if result.version == 0:
            return result
        else:
            logger.warn(
                "Unknown metadata version: %s. The latest version known to "
                "this build of TensorBoard is %s; perhaps a newer build is "
                "available?",
                result.version,
                PROTO_VERSION,
            )
            return result



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import deque
from math import floor, sqrt

import numpy as np
import tensorflow as tf

from tensorboard.plugins.beholder import im_util
from tensorboard.plugins.beholder.shared_config import (
    SECTION_HEIGHT,
    IMAGE_WIDTH,
    DEFAULT_CONFIG,
    SECTION_INFO_FILENAME,
)
from tensorboard.plugins.beholder.file_system_tools import write_pickle

MIN_SQUARE_SIZE = 3


class Visualizer(object):
    def __init__(self, logdir):
        self.logdir = logdir
        self.sections_over_time = deque([], DEFAULT_CONFIG["window_size"])
        self.config = dict(DEFAULT_CONFIG)
        self.old_config = dict(DEFAULT_CONFIG)

    def _reshape_conv_array(self, array, section_height, image_width):
        """Reshape a rank 4 array to be rank 2, where each column of
        block_width is a filter, and each row of block height is an input
        channel. For example:

        [[[[ 11,  21,  31,  41],
           [ 51,  61,  71,  81],
           [ 91, 101, 111, 121]],
          [[ 12,  22,  32,  42],
           [ 52,  62,  72,  82],
           [ 92, 102, 112, 122]],
          [[ 13,  23,  33,  43],
           [ 53,  63,  73,  83],
           [ 93, 103, 113, 123]]],
         [[[ 14,  24,  34,  44],
           [ 54,  64,  74,  84],
           [ 94, 104, 114, 124]],
          [[ 15,  25,  35,  45],
           [ 55,  65,  75,  85],
           [ 95, 105, 115, 125]],
          [[ 16,  26,  36,  46],
           [ 56,  66,  76,  86],
           [ 96, 106, 116, 126]]],
         [[[ 17,  27,  37,  47],
           [ 57,  67,  77,  87],
           [ 97, 107, 117, 127]],
          [[ 18,  28,  38,  48],
           [ 58,  68,  78,  88],
           [ 98, 108, 118, 128]],
          [[ 19,  29,  39,  49],
           [ 59,  69,  79,  89],
           [ 99, 109, 119, 129]]]]

           should be reshaped to:

           [[ 11,  12,  13,  21,  22,  23,  31,  32,  33,  41,  42,  43],
            [ 14,  15,  16,  24,  25,  26,  34,  35,  36,  44,  45,  46],
            [ 17,  18,  19,  27,  28,  29,  37,  38,  39,  47,  48,  49],
            [ 51,  52,  53,  61,  62,  63,  71,  72,  73,  81,  82,  83],
            [ 54,  55,  56,  64,  65,  66,  74,  75,  76,  84,  85,  86],
            [ 57,  58,  59,  67,  68,  69,  77,  78,  79,  87,  88,  89],
            [ 91,  92,  93, 101, 102, 103, 111, 112, 113, 121, 122, 123],
            [ 94,  95,  96, 104, 105, 106, 114, 115, 116, 124, 125, 126],
            [ 97,  98,  99, 107, 108, 109, 117, 118, 119, 127, 128, 129]]
        """

        # E.g. [100, 24, 24, 10]: this shouldn't be reshaped like normal.
        if (
            array.shape[1] == array.shape[2]
            and array.shape[0] != array.shape[1]
        ):
            array = np.rollaxis(np.rollaxis(array, 2), 2)

        block_height, block_width, in_channels = array.shape[:3]
        rows = []

        max_element_count = section_height * int(image_width / MIN_SQUARE_SIZE)
        element_count = 0

        for i in range(in_channels):
            rows.append(array[:, :, i, :].reshape(block_height, -1, order="F"))

            # This line should be left in this position. Gives it one extra row.
            if (
                element_count >= max_element_count
                and not self.config["show_all"]
            ):
                break

            element_count += block_height * in_channels * block_width

        return np.vstack(rows)

    def _reshape_irregular_array(self, array, section_height, image_width):
        """Reshapes arrays of ranks not in {1, 2, 4}"""
        section_area = section_height * image_width
        flattened_array = np.ravel(array)

        if not self.config["show_all"]:
            flattened_array = flattened_array[
                : int(section_area / MIN_SQUARE_SIZE)
            ]

        cell_count = np.prod(flattened_array.shape)
        cell_area = section_area / cell_count

        cell_side_length = max(1, floor(sqrt(cell_area)))
        row_count = max(1, int(section_height / cell_side_length))
        col_count = int(cell_count / row_count)

        # Reshape the truncated array so that it has the same aspect ratio as
        # the section.

        # Truncate whatever remaining values there are that don't fit. Hopefully
        # it doesn't matter that the last few (< section count) aren't there.
        section = np.reshape(
            flattened_array[: row_count * col_count], (row_count, col_count)
        )

        return section

    def _determine_image_width(self, arrays, show_all):
        final_width = IMAGE_WIDTH

        if show_all:
            for array in arrays:
                rank = len(array.shape)

                if rank == 1:
                    width = len(array)
                elif rank == 2:
                    width = array.shape[1]
                elif rank == 4:
                    width = array.shape[1] * array.shape[3]
                else:
                    width = IMAGE_WIDTH

                if width > final_width:
                    final_width = width

        return final_width

    def _determine_section_height(self, array, show_all):
        rank = len(array.shape)
        height = SECTION_HEIGHT

        if show_all:
            if rank == 1:
                height = SECTION_HEIGHT
            if rank == 2:
                height = max(SECTION_HEIGHT, array.shape[0])
            elif rank == 4:
                height = max(SECTION_HEIGHT, array.shape[0] * array.shape[2])
            else:
                height = max(
                    SECTION_HEIGHT, np.prod(array.shape) // IMAGE_WIDTH
                )

        return height

    def _arrays_to_sections(self, arrays):
        """
        input: unprocessed numpy arrays.
        returns: columns of the size that they will appear in the image, not scaled
            for display. That needs to wait until after variance is computed.
        """
        sections = []
        sections_to_resize_later = {}
        show_all = self.config["show_all"]
        image_width = self._determine_image_width(arrays, show_all)

        for array_number, array in enumerate(arrays):
            rank = len(array.shape)
            section_height = self._determine_section_height(array, show_all)

            if rank == 1:
                section = np.atleast_2d(array)
            elif rank == 2:
                section = array
            elif rank == 4:
                section = self._reshape_conv_array(
                    array, section_height, image_width
                )
            else:
                section = self._reshape_irregular_array(
                    array, section_height, image_width
                )
            # Only calculate variance for what we have to. In some cases (biases),
            # the section is larger than the array, so we don't want to calculate
            # variance for the same value over and over - better to resize later.
            # About a 6-7x speedup for a big network with a big variance window.
            section_size = section_height * image_width
            array_size = np.prod(array.shape)

            if section_size > array_size:
                sections.append(section)
                sections_to_resize_later[array_number] = section_height
            else:
                sections.append(
                    im_util.resize(section, section_height, image_width)
                )

        self.sections_over_time.append(sections)

        if self.config["mode"] == "variance":
            sections = self._sections_to_variance_sections(
                self.sections_over_time
            )

        for array_number, height in sections_to_resize_later.items():
            sections[array_number] = im_util.resize(
                sections[array_number], height, image_width
            )
        return sections

    def _sections_to_variance_sections(self, sections_over_time):
        """Computes the variance of corresponding sections over time.

        Returns:
          a list of np arrays.
        """
        variance_sections = []

        for i in range(len(sections_over_time[0])):
            time_sections = [sections[i] for sections in sections_over_time]
            variance = np.var(time_sections, axis=0)
            variance_sections.append(variance)

        return variance_sections

    def _sections_to_image(self, sections):
        padding_size = 5

        sections = im_util.scale_sections(sections, self.config["scaling"])

        final_stack = [sections[0]]
        padding = np.zeros((padding_size, sections[0].shape[1]))

        for section in sections[1:]:
            final_stack.append(padding)
            final_stack.append(section)

        return np.vstack(final_stack).astype(np.uint8)

    def _maybe_clear_deque(self):
        """Clears the deque if certain parts of the config have changed."""

        for config_item in ["values", "mode", "show_all"]:
            if self.config[config_item] != self.old_config[config_item]:
                self.sections_over_time.clear()
                break

        self.old_config = self.config

        window_size = self.config["window_size"]
        if window_size != self.sections_over_time.maxlen:
            self.sections_over_time = deque(
                self.sections_over_time, window_size
            )

    def _save_section_info(self, arrays, sections):
        infos = []

        if self.config["values"] == "trainable_variables":
            names = [x.name for x in tf.compat.v1.trainable_variables()]
        else:
            names = range(len(arrays))

        for array, section, name in zip(arrays, sections, names):
            info = {}

            info["name"] = name
            info["shape"] = str(array.shape)
            info["min"] = "{:.3e}".format(section.min())
            info["mean"] = "{:.3e}".format(section.mean())
            info["max"] = "{:.3e}".format(section.max())
            info["range"] = "{:.3e}".format(section.max() - section.min())
            info["height"] = section.shape[0]

            infos.append(info)

        write_pickle(infos, "{}/{}".format(self.logdir, SECTION_INFO_FILENAME))

    def build_frame(self, arrays):
        self._maybe_clear_deque()

        arrays = arrays if isinstance(arrays, list) else [arrays]

        sections = self._arrays_to_sections(arrays)
        self._save_section_info(arrays, sections)
        final_image = self._sections_to_image(sections)
        final_image = im_util.apply_colormap(
            final_image, self.config["colormap"]
        )

        return final_image

    def update(self, config):
        self.config = config



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""TensorBoard WSGI Application Logic.

Provides TensorBoardWSGIApp for building a TensorBoard WSGI app.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import atexit
import base64
import collections
import contextlib
import hashlib
import json
import os
import re
import shutil
import sqlite3
import tempfile
import textwrap
import threading
import time

import six
from six.moves.urllib import (
    parse as urlparse,
)  # pylint: disable=wrong-import-order

from werkzeug import wrappers

from tensorboard import errors
from tensorboard import plugin_util
from tensorboard.backend import empty_path_redirect
from tensorboard.backend import experiment_id
from tensorboard.backend import experimental_plugin
from tensorboard.backend import http_util
from tensorboard.backend import path_prefix
from tensorboard.backend import security_validator
from tensorboard.backend.event_processing import db_import_multiplexer
from tensorboard.backend.event_processing import (
    data_provider as event_data_provider,
)
from tensorboard.backend.event_processing import (
    plugin_event_accumulator as event_accumulator,
)
from tensorboard.backend.event_processing import (
    plugin_event_multiplexer as event_multiplexer,
)
from tensorboard.plugins import base_plugin
from tensorboard.plugins.audio import metadata as audio_metadata
from tensorboard.plugins.core import core_plugin
from tensorboard.plugins.histogram import metadata as histogram_metadata
from tensorboard.plugins.image import metadata as image_metadata
from tensorboard.plugins.pr_curve import metadata as pr_curve_metadata
from tensorboard.plugins.scalar import metadata as scalar_metadata
from tensorboard.util import tb_logging


DEFAULT_SIZE_GUIDANCE = {
    event_accumulator.TENSORS: 10,
}

# TODO(@wchargin): Once SQL mode is in play, replace this with an
# alternative that does not privilege first-party plugins.
DEFAULT_TENSOR_SIZE_GUIDANCE = {
    scalar_metadata.PLUGIN_NAME: 1000,
    image_metadata.PLUGIN_NAME: 10,
    audio_metadata.PLUGIN_NAME: 10,
    histogram_metadata.PLUGIN_NAME: 500,
    pr_curve_metadata.PLUGIN_NAME: 100,
}

DATA_PREFIX = "/data"
PLUGIN_PREFIX = "/plugin"
PLUGINS_LISTING_ROUTE = "/plugins_listing"
PLUGIN_ENTRY_ROUTE = "/plugin_entry.html"

EXPERIMENTAL_PLUGINS_QUERY_PARAM = "experimentalPlugin"

# Slashes in a plugin name could throw the router for a loop. An empty
# name would be confusing, too. To be safe, let's restrict the valid
# names as follows.
_VALID_PLUGIN_RE = re.compile(r"^[A-Za-z0-9_-]+$")

logger = tb_logging.get_logger()


def _parse_samples_per_plugin(flags):
    result = {}
    if not flags or not flags.samples_per_plugin:
        return result
    for token in flags.samples_per_plugin.split(","):
        k, v = token.strip().split("=")
        result[k] = int(v)
    return result


def _apply_tensor_size_guidance(sampling_hints):
    """Apply user per-summary size guidance overrides."""
    tensor_size_guidance = dict(DEFAULT_TENSOR_SIZE_GUIDANCE)
    tensor_size_guidance.update(sampling_hints)
    return tensor_size_guidance


def standard_tensorboard_wsgi(flags, plugin_loaders, assets_zip_provider):
    """Construct a TensorBoardWSGIApp with standard plugins and multiplexer.

    Args:
      flags: An argparse.Namespace containing TensorBoard CLI flags.
      plugin_loaders: A list of TBLoader instances.
      assets_zip_provider: See TBContext documentation for more information.

    Returns:
      The new TensorBoard WSGI application.

    :type plugin_loaders: list[base_plugin.TBLoader]
    :rtype: TensorBoardWSGI
    """
    data_provider = None
    multiplexer = None
    reload_interval = flags.reload_interval
    if flags.db_import:
        # DB import mode.
        db_uri = flags.db
        # Create a temporary DB file if we weren't given one.
        if not db_uri:
            tmpdir = tempfile.mkdtemp(prefix="tbimport")
            atexit.register(shutil.rmtree, tmpdir)
            db_uri = "sqlite:%s/tmp.sqlite" % tmpdir
        db_connection_provider = create_sqlite_connection_provider(db_uri)
        logger.info("Importing logdir into DB at %s", db_uri)
        multiplexer = db_import_multiplexer.DbImportMultiplexer(
            db_uri=db_uri,
            db_connection_provider=db_connection_provider,
            purge_orphaned_data=flags.purge_orphaned_data,
            max_reload_threads=flags.max_reload_threads,
        )
    elif flags.db:
        # DB read-only mode, never load event logs.
        reload_interval = -1
        db_connection_provider = create_sqlite_connection_provider(flags.db)
        multiplexer = _DbModeMultiplexer(flags.db, db_connection_provider)
    else:
        # Regular logdir loading mode.
        sampling_hints = _parse_samples_per_plugin(flags)
        multiplexer = event_multiplexer.EventMultiplexer(
            size_guidance=DEFAULT_SIZE_GUIDANCE,
            tensor_size_guidance=_apply_tensor_size_guidance(sampling_hints),
            purge_orphaned_data=flags.purge_orphaned_data,
            max_reload_threads=flags.max_reload_threads,
            event_file_active_filter=_get_event_file_active_filter(flags),
        )
        data_provider = event_data_provider.MultiplexerDataProvider(
            multiplexer, flags.logdir or flags.logdir_spec
        )

    if reload_interval >= 0:
        # We either reload the multiplexer once when TensorBoard starts up, or we
        # continuously reload the multiplexer.
        if flags.logdir:
            path_to_run = {os.path.expanduser(flags.logdir): None}
        else:
            path_to_run = parse_event_files_spec(flags.logdir_spec)
        start_reloading_multiplexer(
            multiplexer, path_to_run, reload_interval, flags.reload_task
        )
    return TensorBoardWSGIApp(
        flags, plugin_loaders, data_provider, assets_zip_provider, multiplexer
    )


def _handling_errors(wsgi_app):
    def wrapper(*args):
        (environ, start_response) = (args[-2], args[-1])
        try:
            return wsgi_app(*args)
        except errors.PublicError as e:
            request = wrappers.Request(environ)
            error_app = http_util.Respond(
                request, str(e), "text/plain", code=e.http_code
            )
            return error_app(environ, start_response)
        # Let other exceptions be handled by the server, as an opaque
        # internal server error.

    return wrapper


def TensorBoardWSGIApp(
    flags,
    plugins,
    data_provider=None,
    assets_zip_provider=None,
    deprecated_multiplexer=None,
):
    """Constructs a TensorBoard WSGI app from plugins and data providers.

    Args:
      flags: An argparse.Namespace containing TensorBoard CLI flags.
      plugins: A list of plugin loader instances.
      assets_zip_provider: See TBContext documentation for more information.
      data_provider: Instance of `tensorboard.data.provider.DataProvider`. May
          be `None` if `flags.generic_data` is set to `"false"` in which case
          `deprecated_multiplexer` must be passed instead.
      deprecated_multiplexer: Optional `plugin_event_multiplexer.EventMultiplexer`
          to use for any plugins not yet enabled for the DataProvider API.
          Required if the data_provider argument is not passed.

    Returns:
      A WSGI application that implements the TensorBoard backend.

    :type plugins: list[base_plugin.TBLoader]
    """
    db_uri = None
    db_connection_provider = None
    if isinstance(
        deprecated_multiplexer,
        (db_import_multiplexer.DbImportMultiplexer, _DbModeMultiplexer),
    ):
        db_uri = deprecated_multiplexer.db_uri
        db_connection_provider = deprecated_multiplexer.db_connection_provider
    plugin_name_to_instance = {}
    context = base_plugin.TBContext(
        data_provider=data_provider,
        db_connection_provider=db_connection_provider,
        db_uri=db_uri,
        flags=flags,
        logdir=flags.logdir,
        multiplexer=deprecated_multiplexer,
        assets_zip_provider=assets_zip_provider,
        plugin_name_to_instance=plugin_name_to_instance,
        sampling_hints=_parse_samples_per_plugin(flags),
        window_title=flags.window_title,
    )
    tbplugins = []
    experimental_plugins = []
    for plugin_spec in plugins:
        loader = make_plugin_loader(plugin_spec)
        try:
            plugin = loader.load(context)
        except Exception:
            logger.error(
                "Failed to load plugin %s; ignoring it.",
                getattr(loader.load, "__qualname__", loader.load),
                exc_info=True,
            )
            plugin = None
        if plugin is None:
            continue
        tbplugins.append(plugin)
        if isinstance(
            loader, experimental_plugin.ExperimentalPlugin
        ) or isinstance(plugin, experimental_plugin.ExperimentalPlugin):
            experimental_plugins.append(plugin.plugin_name)
        plugin_name_to_instance[plugin.plugin_name] = plugin
    return TensorBoardWSGI(
        tbplugins, flags.path_prefix, data_provider, experimental_plugins
    )


class TensorBoardWSGI(object):
    """The TensorBoard WSGI app that delegates to a set of TBPlugin."""

    def __init__(
        self,
        plugins,
        path_prefix="",
        data_provider=None,
        experimental_plugins=None,
    ):
        """Constructs TensorBoardWSGI instance.

        Args:
          plugins: A list of base_plugin.TBPlugin subclass instances.
          flags: An argparse.Namespace containing TensorBoard CLI flags.
          data_provider: `tensorboard.data.provider.DataProvider` or
            `None`; if present, will inform the "active" state of
            `/plugins_listing`.
          experimental_plugins: A list of plugin names that are only provided
              experimentally. The corresponding plugins will only be activated for
              a user if the user has specified the plugin with the experimentalPlugin
              query parameter in the URL.

        Returns:
          A WSGI application for the set of all TBPlugin instances.

        Raises:
          ValueError: If some plugin has no plugin_name
          ValueError: If some plugin has an invalid plugin_name (plugin
              names must only contain [A-Za-z0-9_.-])
          ValueError: If two plugins have the same plugin_name
          ValueError: If some plugin handles a route that does not start
              with a slash

        :type plugins: list[base_plugin.TBPlugin]
        """
        self._plugins = plugins
        self._path_prefix = path_prefix
        self._data_provider = data_provider
        self._experimental_plugins = frozenset(experimental_plugins or ())
        if self._path_prefix.endswith("/"):
            # Should have been fixed by `fix_flags`.
            raise ValueError(
                "Trailing slash in path prefix: %r" % self._path_prefix
            )

        self.exact_routes = {
            # TODO(@chihuahua): Delete this RPC once we have skylark rules that
            # obviate the need for the frontend to determine which plugins are
            # active.
            DATA_PREFIX + PLUGINS_LISTING_ROUTE: self._serve_plugins_listing,
            DATA_PREFIX + PLUGIN_ENTRY_ROUTE: self._serve_plugin_entry,
        }
        unordered_prefix_routes = {}

        # Serve the routes from the registered plugins using their name as the route
        # prefix. For example if plugin z has two routes /a and /b, they will be
        # served as /data/plugin/z/a and /data/plugin/z/b.
        plugin_names_encountered = set()
        for plugin in self._plugins:
            if plugin.plugin_name is None:
                raise ValueError("Plugin %s has no plugin_name" % plugin)
            if not _VALID_PLUGIN_RE.match(plugin.plugin_name):
                raise ValueError(
                    "Plugin %s has invalid name %r"
                    % (plugin, plugin.plugin_name)
                )
            if plugin.plugin_name in plugin_names_encountered:
                raise ValueError(
                    "Duplicate plugins for name %s" % plugin.plugin_name
                )
            plugin_names_encountered.add(plugin.plugin_name)

            try:
                plugin_apps = plugin.get_plugin_apps()
            except Exception as e:  # pylint: disable=broad-except
                if (
                    type(plugin) is core_plugin.CorePlugin
                ):  # pylint: disable=unidiomatic-typecheck
                    raise
                logger.warn(
                    "Plugin %s failed. Exception: %s",
                    plugin.plugin_name,
                    str(e),
                )
                continue
            for route, app in plugin_apps.items():
                if not route.startswith("/"):
                    raise ValueError(
                        "Plugin named %r handles invalid route %r: "
                        "route does not start with a slash"
                        % (plugin.plugin_name, route)
                    )
                if (
                    type(plugin) is core_plugin.CorePlugin
                ):  # pylint: disable=unidiomatic-typecheck
                    path = route
                else:
                    path = (
                        DATA_PREFIX
                        + PLUGIN_PREFIX
                        + "/"
                        + plugin.plugin_name
                        + route
                    )

                if path.endswith("/*"):
                    # Note we remove the '*' but leave the slash in place.
                    path = path[:-1]
                    if "*" in path:
                        # note we re-add the removed * in the format string
                        raise ValueError(
                            "Plugin %r handles invalid route '%s*': Only "
                            "trailing wildcards are supported "
                            "(i.e., `/.../*`)" % (plugin.plugin_name, path)
                        )
                    unordered_prefix_routes[path] = app
                else:
                    if "*" in path:
                        raise ValueError(
                            "Plugin %r handles invalid route %r: Only "
                            "trailing wildcards are supported "
                            "(i.e., `/.../*`)" % (plugin.plugin_name, path)
                        )
                    self.exact_routes[path] = app

        # Wildcard routes will be checked in the given order, so we sort them
        # longest to shortest so that a more specific route will take precedence
        # over a more general one (e.g., a catchall route `/*` should come last).
        self.prefix_routes = collections.OrderedDict(
            sorted(
                six.iteritems(unordered_prefix_routes),
                key=lambda x: len(x[0]),
                reverse=True,
            )
        )

        self._app = self._create_wsgi_app()

    def _create_wsgi_app(self):
        """Apply middleware to create the final WSGI app."""
        app = self._route_request
        app = empty_path_redirect.EmptyPathRedirectMiddleware(app)
        app = experiment_id.ExperimentIdMiddleware(app)
        app = path_prefix.PathPrefixMiddleware(app, self._path_prefix)
        app = security_validator.SecurityValidatorMiddleware(app)
        app = _handling_errors(app)
        return app

    @wrappers.Request.application
    def _serve_plugin_entry(self, request):
        """Serves a HTML for iframed plugin entry point.

        Args:
          request: The werkzeug.Request object.

        Returns:
          A werkzeug.Response object.
        """
        name = request.args.get("name")
        plugins = [
            plugin for plugin in self._plugins if plugin.plugin_name == name
        ]

        if not plugins:
            raise errors.NotFoundError(name)

        if len(plugins) > 1:
            # Technically is not possible as plugin names are unique and is checked
            # by the check on __init__.
            reason = (
                "Plugin invariant error: multiple plugins with name "
                "{name} found: {list}"
            ).format(name=name, list=plugins)
            raise AssertionError(reason)

        plugin = plugins[0]
        module_path = plugin.frontend_metadata().es_module_path
        if not module_path:
            return http_util.Respond(
                request, "Plugin is not module loadable", "text/plain", code=400
            )

        # non-self origin is blocked by CSP but this is a good invariant checking.
        if urlparse.urlparse(module_path).netloc:
            raise ValueError("Expected es_module_path to be non-absolute path")

        module_json = json.dumps("." + module_path)
        script_content = "import({}).then((m) => void m.render());".format(
            module_json
        )
        digest = hashlib.sha256(script_content.encode("utf-8")).digest()
        script_sha = base64.b64encode(digest).decode("ascii")

        html = textwrap.dedent(
            """
            <!DOCTYPE html>
            <head><base href="plugin/{name}/" /></head>
            <body><script type="module">{script_content}</script></body>
            """
        ).format(name=name, script_content=script_content)
        return http_util.Respond(
            request, html, "text/html", csp_scripts_sha256s=[script_sha],
        )

    @wrappers.Request.application
    def _serve_plugins_listing(self, request):
        """Serves an object mapping plugin name to whether it is enabled.

        Args:
          request: The werkzeug.Request object.

        Returns:
          A werkzeug.Response object.
        """
        response = collections.OrderedDict()
        eid = plugin_util.experiment_id(request.environ)
        plugins_with_data = frozenset(
            self._data_provider.list_plugins(eid) or frozenset()
            if self._data_provider is not None
            else frozenset()
        )
        plugins_to_skip = self._experimental_plugins - frozenset(
            request.args.getlist(EXPERIMENTAL_PLUGINS_QUERY_PARAM)
        )
        for plugin in self._plugins:
            if plugin.plugin_name in plugins_to_skip:
                continue

            if (
                type(plugin) is core_plugin.CorePlugin
            ):  # pylint: disable=unidiomatic-typecheck
                # This plugin's existence is a backend implementation detail.
                continue

            is_active = bool(
                frozenset(plugin.data_plugin_names()) & plugins_with_data
            )
            if not is_active:
                try:
                    start = time.time()
                    is_active = plugin.is_active()
                    elapsed = time.time() - start
                    logger.info(
                        "Plugin listing: is_active() for %s took %0.3f seconds",
                        plugin.plugin_name,
                        elapsed,
                    )
                except Exception:
                    is_active = False
                    logger.error(
                        "Plugin listing: is_active() for %s failed (marking inactive)",
                        plugin.plugin_name,
                        exc_info=True,
                    )

            plugin_metadata = plugin.frontend_metadata()
            output_metadata = {
                "disable_reload": plugin_metadata.disable_reload,
                "enabled": is_active,
                # loading_mechanism set below
                "remove_dom": plugin_metadata.remove_dom,
                # tab_name set below
            }

            if plugin_metadata.tab_name is not None:
                output_metadata["tab_name"] = plugin_metadata.tab_name
            else:
                output_metadata["tab_name"] = plugin.plugin_name

            es_module_handler = plugin_metadata.es_module_path
            element_name = plugin_metadata.element_name
            is_ng_component = plugin_metadata.is_ng_component
            if is_ng_component:
                if element_name is not None:
                    raise ValueError(
                        "Plugin %r declared as both Angular built-in and legacy"
                        % plugin.plugin_name
                    )
                if es_module_handler is not None:
                    raise ValueError(
                        "Plugin %r declared as both Angular built-in and iframed"
                        % plugin.plugin_name
                    )
                loading_mechanism = {
                    "type": "NG_COMPONENT",
                }
            elif element_name is not None and es_module_handler is not None:
                logger.error(
                    "Plugin %r declared as both legacy and iframed; skipping",
                    plugin.plugin_name,
                )
                continue
            elif element_name is not None and es_module_handler is None:
                loading_mechanism = {
                    "type": "CUSTOM_ELEMENT",
                    "element_name": element_name,
                }
            elif element_name is None and es_module_handler is not None:
                loading_mechanism = {
                    "type": "IFRAME",
                    "module_path": "".join(
                        [
                            request.script_root,
                            DATA_PREFIX,
                            PLUGIN_PREFIX,
                            "/",
                            plugin.plugin_name,
                            es_module_handler,
                        ]
                    ),
                }
            else:
                # As a compatibility measure (for plugins that we don't
                # control), we'll pull it from the frontend registry for now.
                loading_mechanism = {
                    "type": "NONE",
                }
            output_metadata["loading_mechanism"] = loading_mechanism

            response[plugin.plugin_name] = output_metadata
        return http_util.Respond(request, response, "application/json")

    def __call__(self, environ, start_response):
        """Central entry point for the TensorBoard application.

        This __call__ method conforms to the WSGI spec, so that instances of this
        class are WSGI applications.

        Args:
          environ: See WSGI spec (PEP 3333).
          start_response: See WSGI spec (PEP 3333).
        """
        return self._app(environ, start_response)

    def _route_request(self, environ, start_response):
        """Delegate an incoming request to sub-applications.

        This method supports strict string matching and wildcard routes of a
        single path component, such as `/foo/*`. Other routing patterns,
        like regular expressions, are not supported.

        This is the main TensorBoard entry point before middleware is
        applied. (See `_create_wsgi_app`.)

        Args:
          environ: See WSGI spec (PEP 3333).
          start_response: See WSGI spec (PEP 3333).
        """

        request = wrappers.Request(environ)
        parsed_url = urlparse.urlparse(request.path)
        clean_path = _clean_path(parsed_url.path)

        # pylint: disable=too-many-function-args
        if clean_path in self.exact_routes:
            return self.exact_routes[clean_path](environ, start_response)
        else:
            for path_prefix in self.prefix_routes:
                if clean_path.startswith(path_prefix):
                    return self.prefix_routes[path_prefix](
                        environ, start_response
                    )

            logger.warn("path %s not found, sending 404", clean_path)
            return http_util.Respond(
                request, "Not found", "text/plain", code=404
            )(environ, start_response)
        # pylint: enable=too-many-function-args


def parse_event_files_spec(logdir_spec):
    """Parses `logdir_spec` into a map from paths to run group names.

    The `--logdir_spec` flag format is a comma-separated list of path
    specifications. A path spec looks like 'group_name:/path/to/directory' or
    '/path/to/directory'; in the latter case, the group is unnamed. Group names
    cannot start with a forward slash: /foo:bar/baz will be interpreted as a spec
    with no name and path '/foo:bar/baz'.

    Globs are not supported.

    Args:
      logdir: A comma-separated list of run specifications.
    Returns:
      A dict mapping directory paths to names like {'/path/to/directory': 'name'}.
      Groups without an explicit name are named after their path. If logdir is
      None, returns an empty dict, which is helpful for testing things that don't
      require any valid runs.
    """
    files = {}
    if logdir_spec is None:
        return files
    # Make sure keeping consistent with ParseURI in core/lib/io/path.cc
    uri_pattern = re.compile("[a-zA-Z][0-9a-zA-Z.]*://.*")
    for specification in logdir_spec.split(","):
        # Check if the spec contains group. A spec start with xyz:// is regarded as
        # URI path spec instead of group spec. If the spec looks like /foo:bar/baz,
        # then we assume it's a path with a colon. If the spec looks like
        # [a-zA-z]:\foo then we assume its a Windows path and not a single letter
        # group
        if (
            uri_pattern.match(specification) is None
            and ":" in specification
            and specification[0] != "/"
            and not os.path.splitdrive(specification)[0]
        ):
            # We split at most once so run_name:/path:with/a/colon will work.
            run_name, _, path = specification.partition(":")
        else:
            run_name = None
            path = specification
        if uri_pattern.match(path) is None:
            path = os.path.realpath(os.path.expanduser(path))
        files[path] = run_name
    return files


def start_reloading_multiplexer(
    multiplexer, path_to_run, load_interval, reload_task
):
    """Starts automatically reloading the given multiplexer.

    If `load_interval` is positive, the thread will reload the multiplexer
    by calling `ReloadMultiplexer` every `load_interval` seconds, starting
    immediately. Otherwise, reloads the multiplexer once and never again.

    Args:
      multiplexer: The `EventMultiplexer` to add runs to and reload.
      path_to_run: A dict mapping from paths to run names, where `None` as the run
        name is interpreted as a run name equal to the path.
      load_interval: An integer greater than or equal to 0. If positive, how many
        seconds to wait after one load before starting the next load. Otherwise,
        reloads the multiplexer once and never again (no continuous reloading).
      reload_task: Indicates the type of background task to reload with.

    Raises:
      ValueError: If `load_interval` is negative.
    """
    if load_interval < 0:
        raise ValueError("load_interval is negative: %d" % load_interval)

    def _reload():
        while True:
            start = time.time()
            logger.info("TensorBoard reload process beginning")
            for path, name in six.iteritems(path_to_run):
                multiplexer.AddRunsFromDirectory(path, name)
            logger.info(
                "TensorBoard reload process: Reload the whole Multiplexer"
            )
            multiplexer.Reload()
            duration = time.time() - start
            logger.info(
                "TensorBoard done reloading. Load took %0.3f secs", duration
            )
            if load_interval == 0:
                # Only load the multiplexer once. Do not continuously reload.
                break
            time.sleep(load_interval)

    if reload_task == "process":
        logger.info("Launching reload in a child process")
        import multiprocessing

        process = multiprocessing.Process(target=_reload, name="Reloader")
        # Best-effort cleanup; on exit, the main TB parent process will attempt to
        # kill all its daemonic children.
        process.daemon = True
        process.start()
    elif reload_task in ("thread", "auto"):
        logger.info("Launching reload in a daemon thread")
        thread = threading.Thread(target=_reload, name="Reloader")
        # Make this a daemon thread, which won't block TB from exiting.
        thread.daemon = True
        thread.start()
    elif reload_task == "blocking":
        if load_interval != 0:
            raise ValueError(
                "blocking reload only allowed with load_interval=0"
            )
        _reload()
    else:
        raise ValueError("unrecognized reload_task: %s" % reload_task)


def create_sqlite_connection_provider(db_uri):
    """Returns function that returns SQLite Connection objects.

    Args:
      db_uri: A string URI expressing the DB file, e.g. "sqlite:~/tb.db".

    Returns:
      A function that returns a new PEP-249 DB Connection, which must be closed,
      each time it is called.

    Raises:
      ValueError: If db_uri is not a valid sqlite file URI.
    """
    uri = urlparse.urlparse(db_uri)
    if uri.scheme != "sqlite":
        raise ValueError("Only sqlite DB URIs are supported: " + db_uri)
    if uri.netloc:
        raise ValueError("Can not connect to SQLite over network: " + db_uri)
    if uri.path == ":memory:":
        raise ValueError("Memory mode SQLite not supported: " + db_uri)
    path = os.path.expanduser(uri.path)
    params = _get_connect_params(uri.query)
    # TODO(@jart): Add thread-local pooling.
    return lambda: sqlite3.connect(path, **params)


def _get_connect_params(query):
    params = urlparse.parse_qs(query)
    if any(len(v) > 2 for v in params.values()):
        raise ValueError("DB URI params list has duplicate keys: " + query)
    return {k: json.loads(v[0]) for k, v in params.items()}


def _clean_path(path):
    """Removes a trailing slash from a non-root path.

    Arguments:
      path: The path of a request.

    Returns:
      The route to use to serve the request.
    """
    if path != "/" and path.endswith("/"):
        return path[:-1]
    return path


def _get_event_file_active_filter(flags):
    """Returns a predicate for whether an event file load timestamp is active.

    Returns:
      A predicate function accepting a single UNIX timestamp float argument, or
      None if multi-file loading is not enabled.
    """
    if not flags.reload_multifile:
        return None
    inactive_secs = flags.reload_multifile_inactive_secs
    if inactive_secs == 0:
        return None
    if inactive_secs < 0:
        return lambda timestamp: True
    return lambda timestamp: timestamp + inactive_secs >= time.time()


class _DbModeMultiplexer(event_multiplexer.EventMultiplexer):
    """Shim EventMultiplexer to use when in read-only DB mode.

    In read-only DB mode, the EventMultiplexer is nonfunctional - there is no
    logdir to reload, and the data is all exposed via SQL. This class represents
    the do-nothing EventMultiplexer for that purpose, which serves only as a
    conduit for DB-related parameters.

    The load APIs raise exceptions if called, and the read APIs always
    return empty results.
    """

    def __init__(self, db_uri, db_connection_provider):
        """Constructor for `_DbModeMultiplexer`.

        Args:
          db_uri: A URI to the database file in use.
          db_connection_provider: Provider function for creating a DB connection.
        """
        logger.info("_DbModeMultiplexer initializing for %s", db_uri)
        super(_DbModeMultiplexer, self).__init__()
        self.db_uri = db_uri
        self.db_connection_provider = db_connection_provider
        logger.info("_DbModeMultiplexer done initializing")

    def AddRun(self, path, name=None):
        """Unsupported."""
        raise NotImplementedError()

    def AddRunsFromDirectory(self, path, name=None):
        """Unsupported."""
        raise NotImplementedError()

    def Reload(self):
        """Unsupported."""
        raise NotImplementedError()


def make_plugin_loader(plugin_spec):
    """Returns a plugin loader for the given plugin.

    Args:
      plugin_spec: A TBPlugin subclass, or a TBLoader instance or subclass.

    Returns:
      A TBLoader for the given plugin.

    :type plugin_spec:
      Type[base_plugin.TBPlugin] | Type[base_plugin.TBLoader] |
      base_plugin.TBLoader
    :rtype: base_plugin.TBLoader
    """
    if isinstance(plugin_spec, base_plugin.TBLoader):
        return plugin_spec
    if isinstance(plugin_spec, type):
        if issubclass(plugin_spec, base_plugin.TBLoader):
            return plugin_spec()
        if issubclass(plugin_spec, base_plugin.TBPlugin):
            return base_plugin.BasicLoader(plugin_spec)
    raise TypeError("Not a TBLoader or TBPlugin subclass: %r" % (plugin_spec,))



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

import numpy as np
import tensorflow as tf

from tensorboard.plugins.beholder import im_util
from tensorboard.plugins.beholder.file_system_tools import (
    read_pickle,
    write_pickle,
    write_file,
)
from tensorboard.plugins.beholder.shared_config import (
    PLUGIN_NAME,
    TAG_NAME,
    SUMMARY_FILENAME,
    DEFAULT_CONFIG,
    CONFIG_FILENAME,
    SUMMARY_COLLECTION_KEY_NAME,
)
from tensorboard.plugins.beholder import video_writing
from tensorboard.plugins.beholder.visualizer import Visualizer
from tensorboard.util import tb_logging


logger = tb_logging.get_logger()


class Beholder(object):
    def __init__(self, logdir):
        self.PLUGIN_LOGDIR = logdir + "/plugins/" + PLUGIN_NAME

        self.is_recording = False
        self.video_writer = video_writing.VideoWriter(
            self.PLUGIN_LOGDIR,
            outputs=[
                video_writing.FFmpegVideoOutput,
                video_writing.PNGVideoOutput,
            ],
        )

        self.frame_placeholder = tf.compat.v1.placeholder(
            tf.uint8, [None, None, None]
        )
        self.summary_op = tf.compat.v1.summary.tensor_summary(
            TAG_NAME,
            self.frame_placeholder,
            collections=[SUMMARY_COLLECTION_KEY_NAME],
        )

        self.last_image_shape = []
        self.last_update_time = time.time()
        self.config_last_modified_time = -1
        self.previous_config = dict(DEFAULT_CONFIG)

        if not tf.io.gfile.exists(self.PLUGIN_LOGDIR + "/config.pkl"):
            tf.io.gfile.makedirs(self.PLUGIN_LOGDIR)
            write_pickle(
                DEFAULT_CONFIG,
                "{}/{}".format(self.PLUGIN_LOGDIR, CONFIG_FILENAME),
            )

        self.visualizer = Visualizer(self.PLUGIN_LOGDIR)

    def _get_config(self):
        """Reads the config file from disk or creates a new one."""
        filename = "{}/{}".format(self.PLUGIN_LOGDIR, CONFIG_FILENAME)
        modified_time = os.path.getmtime(filename)

        if modified_time != self.config_last_modified_time:
            config = read_pickle(filename, default=self.previous_config)
            self.previous_config = config
        else:
            config = self.previous_config

        self.config_last_modified_time = modified_time
        return config

    def _write_summary(self, session, frame):
        """Writes the frame to disk as a tensor summary."""
        summary = session.run(
            self.summary_op, feed_dict={self.frame_placeholder: frame}
        )
        path = "{}/{}".format(self.PLUGIN_LOGDIR, SUMMARY_FILENAME)
        write_file(summary, path)

    def _get_final_image(self, session, config, arrays=None, frame=None):
        if config["values"] == "frames":
            if frame is None:
                final_image = im_util.get_image_relative_to_script(
                    "frame-missing.png"
                )
            else:
                frame = frame() if callable(frame) else frame
                final_image = im_util.scale_image_for_display(frame)

        elif config["values"] == "arrays":
            if arrays is None:
                final_image = im_util.get_image_relative_to_script(
                    "arrays-missing.png"
                )
                # TODO: hack to clear the info. Should be cleaner.
                self.visualizer._save_section_info([], [])
            else:
                final_image = self.visualizer.build_frame(arrays)

        elif config["values"] == "trainable_variables":
            arrays = [
                session.run(x) for x in tf.compat.v1.trainable_variables()
            ]
            final_image = self.visualizer.build_frame(arrays)

        if len(final_image.shape) == 2:
            # Map grayscale images to 3D tensors.
            final_image = np.expand_dims(final_image, -1)

        return final_image

    def _enough_time_has_passed(self, FPS):
        """For limiting how often frames are computed."""
        if FPS == 0:
            return False
        else:
            earliest_time = self.last_update_time + (1.0 / FPS)
            return time.time() >= earliest_time

    def _update_frame(self, session, arrays, frame, config):
        final_image = self._get_final_image(session, config, arrays, frame)
        self._write_summary(session, final_image)
        self.last_image_shape = final_image.shape

        return final_image

    def _update_recording(self, frame, config):
        """Adds a frame to the current video output."""
        # pylint: disable=redefined-variable-type
        should_record = config["is_recording"]

        if should_record:
            if not self.is_recording:
                self.is_recording = True
                logger.info(
                    "Starting recording using %s",
                    self.video_writer.current_output().name(),
                )
            self.video_writer.write_frame(frame)
        elif self.is_recording:
            self.is_recording = False
            self.video_writer.finish()
            logger.info("Finished recording")

    # TODO: blanket try and except for production? I don't someone's script to die
    #       after weeks of running because of a visualization.
    def update(self, session, arrays=None, frame=None):
        """Creates a frame and writes it to disk.

        Args:
          arrays: a list of np arrays. Use the "custom" option in the client.
          frame: a 2D np array. This way the plugin can be used for video of any
                 kind, not just the visualization that comes with the plugin.

                 frame can also be a function, which only is evaluated when the
                 "frame" option is selected by the client.
        """
        new_config = self._get_config()

        if self._enough_time_has_passed(self.previous_config["FPS"]):
            self.visualizer.update(new_config)
            self.last_update_time = time.time()
            final_image = self._update_frame(session, arrays, frame, new_config)
            self._update_recording(final_image, new_config)

    ##############################################################################

    @staticmethod
    def gradient_helper(optimizer, loss, var_list=None):
        """A helper to get the gradients out at each step.

        Args:
          optimizer: the optimizer op.
          loss: the op that computes your loss value.

        Returns: the gradient tensors and the train_step op.
        """
        if var_list is None:
            var_list = tf.compat.v1.trainable_variables()

        grads_and_vars = optimizer.compute_gradients(loss, var_list=var_list)
        grads = [pair[0] for pair in grads_and_vars]

        return grads, optimizer.apply_gradients(grads_and_vars)


class BeholderHook(tf.estimator.SessionRunHook):
    """SessionRunHook implementation that runs Beholder every step.

    Convenient when using tf.train.MonitoredSession:
    ```python
    beholder_hook = BeholderHook(LOG_DIRECTORY)
    with MonitoredSession(..., hooks=[beholder_hook]) as sess:
      sess.run(train_op)
    ```
    """

    def __init__(self, logdir):
        """Creates new Hook instance.

        Args:
          logdir: Directory where Beholder should write data.
        """
        self._logdir = logdir
        self.beholder = None

    def begin(self):
        self.beholder = Beholder(self._logdir)

    def after_run(self, run_context, unused_run_values):
        self.beholder.update(run_context.session)
